@@ -39,7 +39,7 @@
 #else
 #define CMD_NAME "icainfo"
 #endif
-#define COPYRIGHT "Copyright IBM Corp. 2007, 2021."
+#define COPYRIGHT "Copyright IBM Corp. 2007, 2022."
 
 #define CELL_SIZE 3
 
@@ -61,6 +61,12 @@ static const unsigned int cpacf_nids[] = {
 	NID_ED25519, NID_ED448, NID_X25519, NID_X448 };
 static const unsigned int cpacf_nids_len = sizeof(cpacf_nids) / sizeof(unsigned int);
 
+#ifdef ICA_FIPS
+static const unsigned int cpacf_fips_nids[] = {
+	NID_X9_62_prime256v1, NID_secp384r1, NID_secp521r1 };
+static const unsigned int cpacf_fips_nids_len = sizeof(cpacf_fips_nids) / sizeof(unsigned int);
+#endif
+
 static int is_msa9(void)
 {
 	unsigned int mech_len, j;
@@ -72,6 +78,11 @@ static int is_msa9(void)
 	}
 
 	pmech_list = malloc(sizeof(libica_func_list_element)*mech_len);
+	if (!pmech_list) {
+		perror("is_msa9: error malloc");
+		return 0;
+	}
+
 	if (ica_get_functionlist(pmech_list, &mech_len) != 0) {
 		perror("get_functionlist: ");
 		free(pmech_list);
@@ -102,6 +113,11 @@ static int online_cca_card(void)
 	}
 
 	pmech_list = malloc(sizeof(libica_func_list_element)*mech_len);
+	if (!pmech_list) {
+		perror("online_cca_card: error malloc");
+		return 0;
+	}
+
 	if (ica_get_functionlist(pmech_list, &mech_len) != 0) {
 		perror("get_functionlist: ");
 		free(pmech_list);
@@ -121,6 +137,82 @@ static int online_cca_card(void)
 	return 0;
 }
 
+int rsa_keylen_supported_by_openssl(unsigned int modulus_bitlength)
+{
+	unsigned char modexpo_public_e[512] = { 0 };
+	unsigned char modexpo_public_n[512] = { 0 };
+	unsigned char crt_private_p[256] = { 0 };
+	unsigned char crt_private_q[256] = { 0 };
+	unsigned char crt_private_dp[256] = { 0 };
+	unsigned char crt_private_dq[256] = { 0 };
+	unsigned char crt_private_inv_q[256] = { 0 };
+	ica_adapter_handle_t ah;
+	ica_rsa_key_mod_expo_t public_key;
+	ica_rsa_key_crt_t private_key;
+	int rc;
+
+	rc = ica_open_adapter(&ah);
+	if (rc != 0)
+		return 0;
+
+	public_key.modulus = modexpo_public_n;
+	public_key.exponent = modexpo_public_e;
+	public_key.key_length = (modulus_bitlength + 7) / 8;
+
+	private_key.p = crt_private_p;
+	private_key.q = crt_private_q;
+	private_key.dp = crt_private_dp;
+	private_key.dq = crt_private_dq;
+	private_key.qInverse = crt_private_inv_q;
+	private_key.key_length = (modulus_bitlength + 7) / 8;
+
+	rc = ica_rsa_key_generate_crt(ah, modulus_bitlength,
+							&public_key, &private_key);
+
+	ica_close_adapter(ah);
+
+	return rc == 0 ? 1 : 0;
+}
+
+int get_rsa_minlen(void)
+{
+	int keylen_array[] = { 57, 512, 1024, 2048, 4096 };
+	size_t i;
+
+	for (i = 0; i < sizeof(keylen_array) / sizeof(int); i++) {
+		if (rsa_keylen_supported_by_openssl(keylen_array[i])) {
+			return keylen_array[i];
+		}
+	}
+
+	return -1;
+}
+
+/**
+ * Print out the minimum and maximum RSA key length. The maximum length is
+ * restricted to 4096 bits by crypto cards. The minimum accepted length in
+ * libica is 57 bits, but the available min length depends on the openssl
+ * version and fips mode.
+ */
+void print_rsa(void)
+{
+	int minlen = get_rsa_minlen();
+
+	if (minlen > 0)
+		printf("RSA key lengths: %d ... 4096 bits.\n", minlen);
+	else
+		printf("Error: cannot determine supported RSA key lengths via openssl.\n");
+
+#ifdef ICA_FIPS
+	printf("Built-in FIPS support: FIPS 140-3 mode %s.\n",
+	    ica_fips_status() & ICA_FIPS_MODE ? "active" : "inactive");
+	if (ica_fips_status() >> 1)
+		printf("FIPS SELF-TEST FAILURE. CHECK THE SYSLOG.\n");
+#else
+	printf("No built-in FIPS support.\n");
+#endif /* ICA_FIPS */
+}
+
 static int num_cpacf_curves(void)
 {
 	if (!is_msa9())
@@ -151,8 +243,10 @@ void append_nid(unsigned int nid, unsigned int flag,
 				sname = "";
 			curve_array[i].sname = (unsigned char *)sname;
 			curve_array[i].flags |= flag;
+#ifndef NO_SW_FALLBACKS
 			if (curve_supported_via_openssl(nid) && sw_fallbacks_implemented(nid))
 				curve_array[i].flags |= ICA_FLAG_SW;
+#endif
 			break;
 		}
 	}
@@ -169,8 +263,10 @@ void add_hw_curves(s390_supported_curves_t *curve_array, size_t array_len,
 		for (j = 0; j < array_len; j++) {
 			if (curve_array[j].nid == nids[i]) {
 				curve_array[j].flags |= flag;
+#ifndef NO_SW_FALLBACKS
 				if (curve_supported_via_openssl(nids[i]) && sw_fallbacks_implemented(nids[i]))
 					curve_array[j].flags |= ICA_FLAG_SW;
+#endif
 				merged = 1;
 				break;
 			}
@@ -182,35 +278,43 @@ void add_hw_curves(s390_supported_curves_t *curve_array, size_t array_len,
 
 void add_curves(s390_supported_curves_t *curve_array, size_t array_len)
 {
+#ifdef ICA_FIPS
+	if (is_msa9()) {
+		if (ica_fips_status() & ICA_FIPS_MODE) {
+			add_hw_curves(curve_array, array_len, cpacf_fips_nids, cpacf_fips_nids_len, ICA_FLAG_SHW);
+		} else {
+			add_hw_curves(curve_array, array_len, cpacf_nids, cpacf_nids_len, ICA_FLAG_SHW);
+			if (online_cca_card())
+				add_hw_curves(curve_array, array_len, cca_nids, cca_nids_len, ICA_FLAG_DHW);
+		}
+	} else {
+		/* No MSA9, ECC only available via CEX cards */
+		if (online_cca_card())
+			add_hw_curves(curve_array, array_len, cca_nids, cca_nids_len, ICA_FLAG_DHW);
+	}
+#else
 	if (online_cca_card())
 		add_hw_curves(curve_array, array_len, cca_nids, cca_nids_len, ICA_FLAG_DHW);
 
 	if (is_msa9())
 		add_hw_curves(curve_array, array_len, cpacf_nids, cpacf_nids_len, ICA_FLAG_SHW);
-}
-
-/**
- * These are nids that are fips approved, but libica has no sw fallbacks
- * implemented for them. If they are supported via hw, we want them in the
- * icainfo output.
- */
-unsigned int fips_override(unsigned int nid)
-{
-	switch (nid) {
-	case NID_ED25519:
-	case NID_ED448:
-	case NID_X25519:
-	case NID_X448:
-		return 1;
-	default:
-		return 0;
-	}
+#endif
 }
 
 void print_ec_curves(void)
 {
 	s390_supported_curves_t *curve_array;
 	unsigned int array_len, array_size, n;
+#ifdef NO_CPACF
+	char *no_shw = "-";
+#else
+	char *no_shw = "no";
+#endif
+#ifdef NO_SW_FALLBACKS
+	char *no_sw = "-";
+#else
+	char *no_sw = "no";
+#endif
 
 	array_len = num_cca_curves() + num_cpacf_curves();
 	array_size = array_len * sizeof(s390_supported_curves_t);
@@ -225,38 +329,24 @@ void print_ec_curves(void)
 	printf("-------------------------------------------------------\n");
 	printf("                 |         hardware        |           \n");
 	printf("        EC curve |   dynamic  |   static   |  software \n");
+	printf("                 |            |  (msa=%02d)  |         \n",ica_get_msa_level());
 	printf("-----------------+------------+------------+-----------\n");
 
 	for (n = 0; n < array_len; n++) {
-#ifdef ICA_FIPS
-		/* In fips mode, only allow openssl-fips supported curves, and curves
-		 * that are fips approved, but have no sw fallback implemented.*/
-		if (curve_array[n].nid != 0 &&
-			((curve_array[n].flags & ICA_FLAG_SW) || (fips_override(curve_array[n].nid)))) {
-#else
 		if (curve_array[n].nid != 0) {
-#endif
 			printf("%16s |    %*s     |    %*s     |    %*s\n",
 				curve_array[n].sname,
 				CELL_SIZE,
 				curve_array[n].flags & ICA_FLAG_DHW ? "yes" : "no",
 				CELL_SIZE,
-#ifdef NO_CPACF
-				"-",
-#else
-				curve_array[n].flags & ICA_FLAG_SHW ? "yes" : "no",
-#endif
+				curve_array[n].flags & ICA_FLAG_SHW ? "yes" : no_shw,
 				CELL_SIZE,
-#if defined(NO_SW_FALLBACKS) || defined(NO_CPACF)
-				"-");
-#else
-				curve_array[n].flags & ICA_FLAG_SW ? "yes" : "no");
-#endif
+				curve_array[n].flags & ICA_FLAG_SW ? "yes" : no_sw);
 		}
 	}
 	printf("-------------------------------------------------------\n");
 #ifdef ICA_FIPS
-	printf("Built-in FIPS support: FIPS mode %s.\n",
+	printf("Built-in FIPS support: FIPS 140-3 mode %s.\n",
 	    ica_fips_status() & ICA_FIPS_MODE ? "active" : "inactive");
 	if (ica_fips_status() >> 1)
 		printf("FIPS SELF-TEST FAILURE. CHECK THE SYSLOG.\n");
@@ -274,6 +364,9 @@ void print_version(void)
 #else
 	printf(CMD_NAME ": libica version " VERSION "\n" COPYRIGHT "\n");
 #endif
+
+	if (strcmp(BUILD_VERSION, DEFAULT_BUILD_VERSION) != 0)
+		printf("build: " BUILD_VERSION "\n");
 }
 
 void print_help(char *cmd)
@@ -285,12 +378,16 @@ void print_help(char *cmd)
 	     "\n" "Options:\n"
 	     " -v, --version        show version information\n"
 	     " -c, --list-curves    list supported EC curves\n"
+	     " -r, --list-rsa       list supported RSA key lengths\n"
+	     " -f, --list-fips-exceptions   show fips exception list\n"
 	     " -h, --help           display this help text\n");
 }
 
-#define getopt_string "qcvh"
+#define getopt_string "qcrfvh"
 static struct option getopt_long_options[] = {
 	{"list-curves", 0, 0, 'c'},
+	{"list-rsa", 0, 0, 'r'},
+	{"list-fips-exceptions", 0, 0, 'f'},
 	{"version", 0, 0, 'v'},
 	{"help", 0, 0, 'h'},
 	{0, 0, 0, 0}
@@ -360,6 +457,113 @@ static struct crypt_pair crypt_map[] = {
 	{NULL,0}
 };
 
+int get_index_in_fips_list(libica_fips_indicator_element *fips_list,
+		unsigned int fips_len, unsigned int algo_id)
+{
+	size_t i;
+
+	for (i = 0; i < fips_len; i++) {
+		if (fips_list[i].mech_mode_id == algo_id)
+			return i;
+	}
+
+	return -1;
+}
+
+int get_index_in_mech_list(libica_func_list_element *mech_list,
+		unsigned int mech_len, unsigned int algo_id)
+{
+	size_t i;
+
+	for (i = 0; i < mech_len; i++) {
+		if (mech_list[i].mech_mode_id == algo_id)
+			return i;
+	}
+
+	return -1;
+}
+
+void print_fips_indicator(void)
+{
+#ifndef ICA_FIPS
+	printf("Built-in FIPS support: FIPS 140-3 mode inactive.\n");
+#else
+	unsigned int fips_len, mech_len, i;
+	int j, k;
+	libica_fips_indicator_element *fips_list = NULL;
+	libica_func_list_element *pmech_list = NULL;
+
+	if (!(ica_fips_status() & ICA_FIPS_MODE)) {
+		printf("Built-in FIPS support: Not running in active fips mode.\n");
+		return;
+	}
+
+	/* Get functionlist */
+	if (ica_get_functionlist(NULL, &mech_len) != 0){
+		perror("get_functionlist: ");
+		goto done;
+	}
+
+	pmech_list = malloc(sizeof(libica_func_list_element)*mech_len);
+	if (!pmech_list) {
+		perror("error malloc");
+		goto done;
+	}
+
+	if (ica_get_functionlist(pmech_list, &mech_len) != 0){
+		perror("get_functionlist: ");
+		goto done;
+	}
+
+	/* Get fips indicator list */
+	if (ica_get_fips_indicator(NULL, &fips_len) != 0){
+		perror("get_fips_indicator: ");
+		goto done;
+	}
+
+	fips_list = malloc(sizeof(libica_fips_indicator_element)*fips_len);
+	if (!fips_list) {
+		perror("error malloc");
+		goto done;
+	}
+
+	if (ica_get_fips_indicator(fips_list, &fips_len) != 0){
+		perror("get_fips_indicator: ");
+		goto done;
+	}
+
+	printf("              FIPS service indicator                  \n");
+	printf("------------------------------------------------------\n");
+	printf("               |      Available but non-approved      \n");
+	printf(" function      |   dynamic  |   static   |  software  \n");
+	printf("---------------+--------------------------------------\n");
+
+	for (i = 0; crypt_map[i].algo_id; i++) {
+		j = get_index_in_fips_list(fips_list, fips_len, crypt_map[i].algo_id);
+		k = get_index_in_mech_list(pmech_list, mech_len, crypt_map[i].algo_id);
+		if (j < 0 || k < 0)
+			continue;
+		if (pmech_list[k].flags != 0 && fips_list[j].fips_approved == 0 &&
+			fips_list[j].fips_override == 1) {
+			printf("%14s |    %*s     |    %*s     |    %*s     \n",
+				crypt_map[i].name,
+				CELL_SIZE,
+				pmech_list[k].flags & ICA_FLAG_DHW ? "yes" : "-",
+				CELL_SIZE,
+				pmech_list[k].flags & ICA_FLAG_SHW ? "yes" : "-",
+				CELL_SIZE,
+				pmech_list[k].flags & ICA_FLAG_SW ? "yes" : "-");
+		}
+	}
+
+	printf("------------------------------------------------------\n");
+
+done:
+	free(fips_list);
+	free(pmech_list);
+	ica_cleanup();
+#endif
+}
 
 int main(int argc, char **argv)
 {
@@ -368,13 +572,30 @@ int main(int argc, char **argv)
 	unsigned int mech_len, j;
 	libica_func_list_element *pmech_list = NULL;
 	unsigned int i;
+	char *no_dhw = "no";
+#ifdef NO_CPACF
+	char *no_shw = "-";
+#else
+	char *no_shw = "no";
+#endif
+#ifdef NO_SW_FALLBACKS
+	char *no_sw = "-";
+#else
+	char *no_sw = "no";
+#endif
 
 	while ((rc = getopt_long(argc, argv, getopt_string,
 				 getopt_long_options, &index)) != -1) {
 		switch (rc) {
+		case 'f':
+			print_fips_indicator();
+			exit(0);
 		case 'c':
 			print_ec_curves();
 			exit(0);
+		case 'r':
+			print_rsa();
+			exit(0);
 		case 'v':
 			print_version();
 			exit(0);
@@ -385,6 +606,7 @@ int main(int argc, char **argv)
 		default:
 			fprintf(stderr, "Try '%s --help' for more"
 				" information.\n", basename(argv[0]));
+			ica_cleanup();
 			exit(1);
 		}
 	}
@@ -392,6 +614,7 @@ int main(int argc, char **argv)
 		fprintf(stderr, "%s: invalid option.\n"
 			"Try '%s --help' for more information.\n",
 			argv[0], basename(argv[0]));
+		ica_cleanup();
 		exit(1);
 	}
 
@@ -400,54 +623,48 @@ int main(int argc, char **argv)
 
 	if (ica_get_functionlist(NULL, &mech_len) != 0){
 		perror("get_functionlist: ");
+		ica_cleanup();
 		return EXIT_FAILURE;
 	}
+
 	pmech_list = malloc(sizeof(libica_func_list_element)*mech_len);
+	if (!pmech_list) {
+		perror("error malloc");
+		return EXIT_FAILURE;
+	}
+
 	if (ica_get_functionlist(pmech_list, &mech_len) != 0){
 		perror("get_functionlist: ");
 		free(pmech_list);
+		ica_cleanup();
 		return EXIT_FAILURE;
 	}
 
 	printf("               |         hardware        |            \n");
 	printf(" function      |   dynamic  |   static   |  software  \n");
+	printf("               |            |  (msa=%02d)  |            \n",ica_get_msa_level());
 	printf("---------------+------------+------------+------------\n");
-	for(i = 0; crypt_map[i].algo_id; i++){
-		for(j = 0; j < mech_len; j++){
-			if(crypt_map[i].algo_id == pmech_list[j].mech_mode_id){
+	for (i = 0; crypt_map[i].algo_id; i++) {
+		for (j = 0; j < mech_len; j++) {
+			if (crypt_map[i].algo_id == pmech_list[j].mech_mode_id) {
 #ifdef ICA_FIPS
-				if (((ica_fips_status() & ICA_FIPS_MODE)
-				    && !fips_approved(pmech_list[j].mech_mode_id))
-				    || ica_fips_status() >> 1) {
-#if defined(NO_SW_FALLBACKS) && defined(NO_CPACF)
-					printf("%14s |  blocked   |      -     |      -   \n",
-#elif defined (NO_CPACF)
-					printf("%14s |  blocked   |      -     |      -   \n",
-#elif defined (NO_SW_FALLBACKS)
-					printf("%14s |  blocked   |   blocked  |      -   \n",
-#else
-					printf("%14s |  blocked   |   blocked  |   blocked\n",
-#endif
+				if (((ica_fips_status() & ICA_FIPS_MODE) &&
+					!fips_approved(pmech_list[j].mech_mode_id) &&
+					!fips_override(pmech_list[j].mech_mode_id)) ||
+					ica_fips_status() >> 1) {
+					printf("%14s |  blocked   |  blocked   |  blocked\n",
 						crypt_map[i].name);
 					break;
 				}
 #endif /* ICA_FIPS */
-				printf("%14s |    %*s     |    %*s     |    %*s\n",
+				printf("%14s |    %*s     |    %*s     |    %*s     \n",
 					crypt_map[i].name,
 					CELL_SIZE,
-					pmech_list[j].flags & ICA_FLAG_DHW ? "yes" : "no",
+					pmech_list[j].flags & ICA_FLAG_DHW ? "yes" : no_dhw,
 					CELL_SIZE,
-#ifdef NO_CPACF
-					pmech_list[j].flags & ICA_FLAG_SHW ? "yes" : "-",
-#else
-					pmech_list[j].flags & ICA_FLAG_SHW ? "yes" : "no",
-#endif
+					pmech_list[j].flags & ICA_FLAG_SHW ? "yes" : no_shw,
 					CELL_SIZE,
-#if defined(NO_SW_FALLBACKS) || defined(NO_CPACF)
-					pmech_list[j].flags & ICA_FLAG_SW ? "yes" : "-");
-#else
-					pmech_list[j].flags & ICA_FLAG_SW ? "yes" : "no");
-#endif
+					pmech_list[j].flags & ICA_FLAG_SW ? "yes" : no_sw);
 			}
 		}
 	}
@@ -455,8 +672,8 @@ int main(int argc, char **argv)
 	printf("------------------------------------------------------\n");
 
 #ifdef ICA_FIPS
-	printf("Built-in FIPS support: FIPS mode %s.\n",
-	    ica_fips_status() & ICA_FIPS_MODE ? "active" : "inactive");
+	printf("Built-in FIPS support: FIPS 140-3 mode %s.\n",
+		ica_fips_status() & ICA_FIPS_MODE ? "active" : "inactive");
 	if (ica_fips_status() >> 1)
 		printf("FIPS SELF-TEST FAILURE. CHECK THE SYSLOG.\n");
 #else
@@ -470,5 +687,6 @@ int main(int argc, char **argv)
 	printf("CPACF support (including fallbacks) is disabled in libica-cex.\n");
 #endif
 
+	ica_cleanup();
 	return EXIT_SUCCESS;
 }
