@@ -36,6 +36,7 @@ unsigned int sha1_switch, sha256_switch, sha512_switch, sha3_switch, des_switch,
 	     prng_switch, tdea128_switch, tdea192_switch, sha512_drng_switch,
 	     msa4_switch, msa5_switch, msa8_switch, trng_switch, msa9_switch,
 		 ecc_via_online_card, any_card_online;
+int msa;
 
 #define CARD_AVAILABLE		0x01
 #define CEXnA_AVAILABLE		0x02
@@ -421,7 +422,7 @@ unsigned int search_for_cards()
 
 void s390_crypto_switches_init(void)
 {
-	int msa, flags;
+	int flags;
 
 	msa = read_facility_bits();
 	if (!msa)
@@ -447,68 +448,71 @@ void s390_crypto_switches_init(void)
  * SW  - software support
  * Bit field flags: [0|0|0|0|0|SHW|DHW|SW]
  * The last field represents the property flags indicating key lengths
+ * Update with libica 4.2.0:
+ *  - Added two more int fields for the fips service indicator. They do not
+ *    affect the libica functionlist.
  */
 libica_func_list_element_int icaList[] = {
- {SHA1,   KIMD, SHA_1  , 0, 0},
- {SHA224, KIMD, SHA_256, 0, 0},
- {SHA256, KIMD, SHA_256, 0, 0},
- {SHA384, KIMD, SHA_512, 0, 0},
- {SHA512, KIMD, SHA_512, 0, 0},
- {SHA512_224, KIMD, SHA_512_224, 0, 0},
- {SHA512_256, KIMD, SHA_512_256, 0, 0},
- {SHA3_224, KIMD, SHA_3_224, 0, 0},
- {SHA3_256, KIMD, SHA_3_256, 0, 0},
- {SHA3_384, KIMD, SHA_3_384, 0, 0},
- {SHA3_512, KIMD, SHA_3_512, 0, 0},
- {SHAKE128, KIMD, SHAKE_128, 0, 0},
- {SHAKE256, KIMD, SHAKE_256, 0, 0},
- {G_HASH, KIMD, GHASH, 0, 0},
+ {SHA1,   KIMD, SHA_1  , 0, 0, 0, 0},
+ {SHA224, KIMD, SHA_256, 0, 0, 0, 0},
+ {SHA256, KIMD, SHA_256, 0, 0, 0, 0},
+ {SHA384, KIMD, SHA_512, 0, 0, 0, 0},
+ {SHA512, KIMD, SHA_512, 0, 0, 0, 0},
+ {SHA512_224, KIMD, SHA_512_224, 0, 0, 0, 0},
+ {SHA512_256, KIMD, SHA_512_256, 0, 0, 0, 0},
+ {SHA3_224, KIMD, SHA_3_224, 0, 0, 0, 0},
+ {SHA3_256, KIMD, SHA_3_256, 0, 0, 0, 0},
+ {SHA3_384, KIMD, SHA_3_384, 0, 0, 0, 0},
+ {SHA3_512, KIMD, SHA_3_512, 0, 0, 0, 0},
+ {SHAKE128, KIMD, SHAKE_128, 0, 0, 0, 0},
+ {SHAKE256, KIMD, SHAKE_256, 0, 0, 0, 0},
+ {G_HASH, KIMD, GHASH, 0, 0, 0, 0},
 
- {DES_ECB,      KMC,  DEA_ENCRYPT, ICA_FLAG_SW, 0},
- {DES_CBC,      KMC,  DEA_ENCRYPT, ICA_FLAG_SW, 0},
- {DES_OFB,      MSA4, DEA_ENCRYPT, 0, 0},
- {DES_CFB,      MSA4, DEA_ENCRYPT, 0, 0},
- {DES_CTR,      MSA4, DEA_ENCRYPT, 0, 0},
- {DES_CMAC,     MSA4, DEA_ENCRYPT, 0, 0}, // CPACF only (MSA4)
+ {DES_ECB,      KMC,  DEA_ENCRYPT, ICA_FLAG_SW, 0, 0, 0},
+ {DES_CBC,      KMC,  DEA_ENCRYPT, ICA_FLAG_SW, 0, 0, 0},
+ {DES_OFB,      MSA4, DEA_ENCRYPT, 0, 0, 0, 0},
+ {DES_CFB,      MSA4, DEA_ENCRYPT, 0, 0, 0, 0},
+ {DES_CTR,      MSA4, DEA_ENCRYPT, 0, 0, 0, 0},
+ {DES_CMAC,     MSA4, DEA_ENCRYPT, 0, 0, 0, 0}, // CPACF only (MSA4)
 
- {DES3_ECB,     KMC,  TDEA_192_ENCRYPT, ICA_FLAG_SW, 0},
- {DES3_CBC,     KMC,  TDEA_192_ENCRYPT, ICA_FLAG_SW, 0},
- {DES3_OFB,     MSA4, TDEA_192_ENCRYPT,           0, 0},
- {DES3_CFB,     MSA4, TDEA_192_ENCRYPT,           0, 0},
- {DES3_CTR,     MSA4, TDEA_192_ENCRYPT,           0, 0},
- {DES3_CMAC,    MSA4, TDEA_192_ENCRYPT,           0, 0},
+ {DES3_ECB,     KMC,  TDEA_192_ENCRYPT, ICA_FLAG_SW, 0, 0, 0},
+ {DES3_CBC,     KMC,  TDEA_192_ENCRYPT, ICA_FLAG_SW, 0, 0, 0},
+ {DES3_OFB,     MSA4, TDEA_192_ENCRYPT,           0, 0, 0, 0},
+ {DES3_CFB,     MSA4, TDEA_192_ENCRYPT,           0, 0, 0, 0},
+ {DES3_CTR,     MSA4, TDEA_192_ENCRYPT,           0, 0, 0, 0},
+ {DES3_CMAC,    MSA4, TDEA_192_ENCRYPT,           0, 0, 0, 0},
 
- {AES_ECB,      KMC,  AES_128_ENCRYPT, ICA_FLAG_SW, 0},
- {AES_CBC,      KMC,  AES_128_ENCRYPT, ICA_FLAG_SW, 0},
- {AES_OFB,      MSA4, AES_128_ENCRYPT,           0, 0},
- {AES_CFB,      MSA4, AES_128_ENCRYPT,           0, 0},
- {AES_CTR,      MSA4, AES_128_ENCRYPT,           0, 0},
- {AES_CMAC,     MSA4, AES_128_ENCRYPT,           0, 0},
- {AES_CCM,      MSA4, AES_128_ENCRYPT,           0, 0},
- {AES_GCM,      MSA4, AES_128_ENCRYPT,           0, 0},
- {AES_GCM_KMA,  MSA8, AES_128_GCM_ENCRYPT,       0, 0},
- {AES_XTS,      MSA4, AES_128_XTS_ENCRYPT,       0, 0},
- {P_RNG,        ADAPTER, 0, ICA_FLAG_SHW | ICA_FLAG_SW, 0}, // SHW (CPACF) + SW
- {EC_DH,        ADAPTER, 0, ICA_FLAG_SW, 0},
- {EC_DSA_SIGN,	ADAPTER, 0, ICA_FLAG_SW, 0},
- {EC_DSA_VERIFY, ADAPTER, 0, ICA_FLAG_SW, 0},
- {EC_KGEN,      ADAPTER, 0, ICA_FLAG_SW, 0},
- {ED25519_KEYGEN, MSA9, SCALAR_MULTIPLY_ED25519, 0, 0},
- {ED25519_SIGN,   MSA9, EDDSA_SIGN_ED25519, 0, 0},
- {ED25519_VERIFY, MSA9, EDDSA_VERIFY_ED25519, 0, 0},
- {ED448_KEYGEN,   MSA9, SCALAR_MULTIPLY_ED448, 0, 0},
- {ED448_SIGN,     MSA9, EDDSA_SIGN_ED448, 0, 0},
- {ED448_VERIFY,   MSA9, EDDSA_VERIFY_ED448, 0, 0},
- {X25519_KEYGEN,   MSA9, SCALAR_MULTIPLY_X25519, 0, 0},
- {X25519_DERIVE,   MSA9, SCALAR_MULTIPLY_X25519, 0, 0},
- {X448_KEYGEN,   MSA9, SCALAR_MULTIPLY_X448, 0, 0},
- {X448_DERIVE,   MSA9, SCALAR_MULTIPLY_X448, 0, 0},
- {RSA_ME,       ADAPTER, 0, ICA_FLAG_SW, 0},
- {RSA_CRT,      ADAPTER, 0, ICA_FLAG_SW, 0},
- {RSA_KEY_GEN_ME, ADAPTER, 0, ICA_FLAG_SW, 0},  // SW (openssl)
- {RSA_KEY_GEN_CRT, ADAPTER, 0, ICA_FLAG_SW, 0}, // SW (openssl)
+ {AES_ECB,      KMC,  AES_128_ENCRYPT, ICA_FLAG_SW, 0, 0, 0},
+ {AES_CBC,      KMC,  AES_128_ENCRYPT, ICA_FLAG_SW, 0, 0, 0},
+ {AES_OFB,      MSA4, AES_128_ENCRYPT,           0, 0, 0, 0},
+ {AES_CFB,      MSA4, AES_128_ENCRYPT,           0, 0, 0, 0},
+ {AES_CTR,      MSA4, AES_128_ENCRYPT,           0, 0, 0, 0},
+ {AES_CMAC,     MSA4, AES_128_ENCRYPT,           0, 0, 0, 0},
+ {AES_CCM,      MSA4, AES_128_ENCRYPT,           0, 0, 0, 0},
+ {AES_GCM,      MSA4, AES_128_ENCRYPT,           0, 0, 0, 0},
+ {AES_GCM_KMA,  MSA8, AES_128_GCM_ENCRYPT,       0, 0, 0, 0},
+ {AES_XTS,      MSA4, AES_128_XTS_ENCRYPT,       0, 0, 0, 0},
+ {P_RNG,        ADAPTER, 0, ICA_FLAG_SHW | ICA_FLAG_SW, 0, 0, 0}, // SHW (CPACF) + SW
+ {EC_DH,        ADAPTER, 0, ICA_FLAG_SW, 0, 0, 0},
+ {EC_DSA_SIGN,	ADAPTER, 0, ICA_FLAG_SW, 0, 0, 0},
+ {EC_DSA_VERIFY, ADAPTER, 0, ICA_FLAG_SW, 0, 0, 0},
+ {EC_KGEN,      ADAPTER, 0, ICA_FLAG_SW, 0, 0, 0},
+ {ED25519_KEYGEN, MSA9, SCALAR_MULTIPLY_ED25519, 0, 0, 0, 0},
+ {ED25519_SIGN,   MSA9, EDDSA_SIGN_ED25519, 0, 0, 0, 0},
+ {ED25519_VERIFY, MSA9, EDDSA_VERIFY_ED25519, 0, 0, 0, 0},
+ {ED448_KEYGEN,   MSA9, SCALAR_MULTIPLY_ED448, 0, 0, 0, 0},
+ {ED448_SIGN,     MSA9, EDDSA_SIGN_ED448, 0, 0, 0, 0},
+ {ED448_VERIFY,   MSA9, EDDSA_VERIFY_ED448, 0, 0, 0, 0},
+ {X25519_KEYGEN,   MSA9, SCALAR_MULTIPLY_X25519, 0, 0, 0, 0},
+ {X25519_DERIVE,   MSA9, SCALAR_MULTIPLY_X25519, 0, 0, 0, 0},
+ {X448_KEYGEN,   MSA9, SCALAR_MULTIPLY_X448, 0, 0, 0, 0},
+ {X448_DERIVE,   MSA9, SCALAR_MULTIPLY_X448, 0, 0, 0, 0},
+ {RSA_ME,       ADAPTER, 0, ICA_FLAG_SW, 0, 0, 0},
+ {RSA_CRT,      ADAPTER, 0, ICA_FLAG_SW, 0, 0, 0},
+ {RSA_KEY_GEN_ME, ADAPTER, 0, ICA_FLAG_SW, 0, 0, 0},  // SW (openssl)
+ {RSA_KEY_GEN_CRT, ADAPTER, 0, ICA_FLAG_SW, 0, 0, 0}, // SW (openssl)
 
- {SHA512_DRNG, PPNO, SHA512_DRNG_GEN, ICA_FLAG_SW, 0},
+ {SHA512_DRNG, PPNO, SHA512_DRNG_GEN, ICA_FLAG_SW, 0, 0, 0},
 
 };
 
@@ -562,6 +566,9 @@ int s390_initialize_functionlist()
 			break;
 		case MSA8:
 			e->flags |= *s390_kma_functions[e->id].enabled ? ICA_FLAG_SHW : 0;
+			e->property |= ICA_PROPERTY_AES_128;
+			e->property |= ICA_PROPERTY_AES_192;
+			e->property |= ICA_PROPERTY_AES_256;
 			break;
 		case MSA9:
 			if (e->mech_mode_id == ED25519_KEYGEN
@@ -585,11 +592,13 @@ int s390_initialize_functionlist()
 		case EC_DSA_SIGN: /* fall-through */
 		case EC_DSA_VERIFY: /* fall-through */
 		case EC_KGEN:
+			e->flags |= *s390_kdsa_functions[e->id].enabled ? ICA_FLAG_SHW : 0;
+			if (e->flags)
+				e->property |= ICA_PROPERTY_EC_NIST;
 			if (ecc_via_online_card) {
 				e->flags |= ICA_FLAG_DHW;
 				e->property |= ICA_PROPERTY_EC_BP | ICA_PROPERTY_EC_NIST;
 			}
-			e->flags |= *s390_kdsa_functions[e->id].enabled ? ICA_FLAG_SHW : 0;
 			break;
 		case RSA_ME: /* fall-through */
 		case RSA_CRT:
@@ -600,12 +609,10 @@ int s390_initialize_functionlist()
 			break;
 		case RSA_KEY_GEN_ME: /* fall-through */
 		case RSA_KEY_GEN_CRT:
-#if defined(ICA_FIPS) && OPENSSL_VERSION_PREREQ(3, 0)
-			if (fips & ICA_FIPS_MODE)
-				e->property |= ICA_PROPERTY_RSA_FIPS;
-			else
+			if (any_card_online) {
+				/* sw flag already pre-set in icaList */
 				e->property |= ICA_PROPERTY_RSA_ALL;
-#endif
+			}
 			break;
 		default:
 			/* Do nothing. */
@@ -670,25 +677,113 @@ int s390_get_functionlist(libica_func_list_element *pmech_list,
 	 *   approved.
 	 * - We are in an error state.
 	 * */
-	if (((fips & ICA_FIPS_MODE) && !fips_approved(icaList[x].mech_mode_id))
+	if (((fips & ICA_FIPS_MODE) && !fips_approved(pmech_list[x].mech_mode_id &&
+		!fips_override(pmech_list[x].mech_mode_id)))
 	    || fips >> 1) {
 		pmech_list[x].flags = 0;
 		pmech_list[x].property = 0;
 	}
+
+	/* Adjust the flags and properties for algorithms that are allowed in fips
+	 * mode, but not on any hardware, with any openssl, not with any key
+	 * length, curve etc. */
+	if (fips & ICA_FIPS_MODE) {
+		/* RSA >= 2048 bits in FIPS 140-3 mode */
+		switch (pmech_list[x].mech_mode_id) {
+		case RSA_KEY_GEN_ME:
+		case RSA_KEY_GEN_CRT:
+		case RSA_ME:
+		case RSA_CRT:
+			if (pmech_list[x].flags)
+				pmech_list[x].property = ICA_PROPERTY_RSA_FIPS;
+			else
+				pmech_list[x].property = 0;
+			break;
+		default:
+			break;
+		}
+
+		/* ECDSA/ECDH in FIPS 140-3 mode */
+		switch (pmech_list[x].mech_mode_id) {
+		case EC_KGEN:
+			/* EC keygen only via openssl because of internal key checks */
+			pmech_list[x].flags = ICA_FLAG_SW;
+			pmech_list[x].property &= ~ICA_PROPERTY_EC_BP;
+			pmech_list[x].property &= ~ICA_PROPERTY_EC_ED;
+			break;
+		case EC_DH:
+		case EC_DSA_SIGN:
+		case EC_DSA_VERIFY:
+			/* EC sign/verify/dh only NIST curves via CPACF because of self-tests */
+			pmech_list[x].flags &= ~ICA_FLAG_DHW;
+			pmech_list[x].property &= ~ICA_PROPERTY_EC_BP;
+			pmech_list[x].property &= ~ICA_PROPERTY_EC_ED;
+			break;
+		default:
+			break;
+		}
+	}
 #endif /* ICA_FIPS */
 
 #ifdef NO_CPACF
 	pmech_list[x].flags &= ~ICA_FLAG_SHW;
 	/* NO_CPACF also removes sw fallbacks for CPACF based functions, but not
-	 * for card-based functions. */
-	if (pmech_list[x].flags & ICA_FLAG_SW && !(pmech_list[x].flags & ICA_FLAG_DHW)) {
+	 * for card-based functions and RSA keygen, which is only available in sw. */
+	if (pmech_list[x].flags & ICA_FLAG_SW && !(pmech_list[x].flags & ICA_FLAG_DHW) &&
+		pmech_list[x].mech_mode_id != RSA_KEY_GEN_ME &&
+		pmech_list[x].mech_mode_id != RSA_KEY_GEN_CRT) {
 		pmech_list[x].flags &= ~ICA_FLAG_SW;
 	}
 #endif /* NO_CPACF */
 
 #ifdef NO_SW_FALLBACKS
-	pmech_list[x].flags &= ~ICA_FLAG_SW;
+	/* Set SW flag to 0 if we don't have sw fallbacks, except for RSA keygen,
+	 * because there is no hw path for RSA keygen. */
+	if (pmech_list[x].mech_mode_id != RSA_KEY_GEN_ME &&
+		pmech_list[x].mech_mode_id != RSA_KEY_GEN_CRT)
+		pmech_list[x].flags &= ~ICA_FLAG_SW;
+#endif /* NO_SW_FALLBACKS */
+
+#ifdef ICA_FIPS
+	if (fips & ICA_FIPS_MODE) {
+		/* When running in fips mode, finally enable EC keygen via sw only.
+		 * This was already set above, but the NO_SW_FALLBACKS block might
+		 * have toggled the sw flag off. The property field is already set. */
+		if (pmech_list[x].mech_mode_id == EC_KGEN)
+			pmech_list[x].flags = ICA_FLAG_SW;
+	}
 #endif
+
+	/* Finally zeroize properties if no flags set, i.e. function not available */
+	if (pmech_list[x].flags == 0)
+		pmech_list[x].property = 0;
   }
   return 0;
 }
+
+#ifdef ICA_FIPS
+int s390_get_fips_indicator(libica_fips_indicator_element *indicator_list,
+						unsigned int *indicator_list_len)
+{
+	size_t i;
+
+	if (!indicator_list_len)
+		return EINVAL;
+
+	if (!indicator_list) {
+		*indicator_list_len = sizeof(icaList) / sizeof(libica_fips_indicator_element);
+		return 0;
+	}
+
+	if (*indicator_list_len < (sizeof(icaList) / sizeof(libica_fips_indicator_element)))
+		return EINVAL;
+
+	for (i = 0; i < *indicator_list_len; i++) {
+		indicator_list[i].mech_mode_id = icaList[i].mech_mode_id;
+		indicator_list[i].fips_approved = fips_approved(icaList[i].mech_mode_id);
+		indicator_list[i].fips_override = fips_override(icaList[i].mech_mode_id);
+	}
+
+	return 0;
+}
+#endif /* ICA_FIPS */
